-- Lean version: 3.4.2
-- Mathlib version: 2019-07-31

import tactic.push_neg
import tactic.rcases

def support (P : (string → ℕ) → Prop) : set string :=
{ a | ∃ (s1 s2 : string → ℕ), (∀ a', a' ≠ a → s1 a' = s2 a') ∧ P s1 ≠ P s2 }

/--
Given a function `f : string → ℕ`, `(a a' : string)`, and `b : ℕ`, `(subst f a b) a'` returns `b` if `a' = a`
and `f a'` otherwise.
-/
@[simp]
noncomputable def subst (f : string → ℕ) (a : string) (b : ℕ) :=
(λ a', if a' = a then b else f a')

-- `↦` can be typed by `\mapsto`
notation f `[`a` ↦ `b`]` := subst f a b

lemma xams_lemma (P : (string → ℕ) → Prop) {a : string} :
  a ∉ support P → (∀ s b, P (s[a ↦ b]) = P s) :=
begin
  assume (hyp : a ∉ support P) s b,
  change ¬∃ (s1 s2 : string → ℕ), (∀ (a' : string), a' ≠ a → s1 a' = s2 a') ∧ P s1 ≠ P s2 at hyp,
  replace hyp : ∀ (s1 s2 : string → ℕ), (∃ (a' : string), a' ≠ a ∧ s1 a' ≠ s2 a') ∨ P s1 = P s2, by
    { push_neg at hyp, exact hyp },
  have : (∃ (a' : string), a' ≠ a ∧ s[a ↦ b] a' ≠ s a') ∨ P (s[a ↦ b]) = P s, from hyp (s[a ↦ b]) s,
  cases this,
  { rcases this with ⟨a', a'_ne_a, _⟩,
    have : s[a ↦ b] a' = s a', by simp [a'_ne_a],
    contradiction },
  { assumption }
end
